# Copyright (C) 2013-2014 Science and Technology Facilities Council.
# Copyright (C) 2017 East Asian Observatory.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

# If running under Python 2, import the itertools.izip function.  In
# Python 3 this is not necessary -- the import will fail and we use the
# normal zip function instead.
try:
    from itertools import izip
except ImportError as e:
    izip = zip

from astropy.io import fits
from datetime import datetime
from math import log
import numpy as np

from ..version import version


def write_moc_fits_hdu(moc):
    """Create a FITS table HDU representation of a MOC.
    """

    # Ensure data are normalized.
    moc.normalize()

    # Determine whether a 32 or 64 bit column is required.
    if moc.order < 14:
        moc_type = np.int32
        col_type = 'J'
    else:
        moc_type = np.int64
        col_type = 'K'

    # Convert to the NUNIQ value which guarantees that one of the
    # top two bits is set so that the order of the value can be
    # determined.
    nuniq = []
    for (order, cells) in moc:
        uniq_prefix = 4 * (4 ** order)
        for npix in cells:
            nuniq.append(npix + uniq_prefix)

    # Prepare the data, and sort into numerical order.
    nuniq = np.array(nuniq, dtype=moc_type)
    nuniq.sort()

    # Create the FITS file.
    col = fits.Column(name='UNIQ', format=col_type, array=nuniq)

    cols = fits.ColDefs([col])
    rec = fits.FITS_rec.from_columns(cols)
    tbhdu = fits.BinTableHDU(rec)

    # Mandatory Keywords.
    tbhdu.header['PIXTYPE'] = 'HEALPIX'
    tbhdu.header['ORDERING'] = 'NUNIQ'
    tbhdu.header['COORDSYS'] = 'C'
    tbhdu.header['MOCORDER'] = moc.order
    tbhdu.header.comments['PIXTYPE'] = 'HEALPix magic code'
    tbhdu.header.comments['ORDERING'] = 'NUNIQ coding method'
    tbhdu.header.comments['COORDSYS'] = 'ICRS reference frame'
    tbhdu.header.comments['MOCORDER'] = 'MOC resolution (best order)'

    # Optional Keywords.
    tbhdu.header['MOCTOOL'] = 'PyMOC ' + version
    tbhdu.header.comments['MOCTOOL'] = 'Name of MOC generator'
    if moc.type is not None:
        tbhdu.header['MOCTYPE'] = moc.type
        tbhdu.header.comments['MOCTYPE'] = 'Source type (IMAGE or CATALOG)'
    if moc.id is not None:
        tbhdu.header['MOCID'] = moc.id
        tbhdu.header.comments['MOCID'] = 'Identifier of the collection'
    if moc.origin is not None:
        tbhdu.header['ORIGIN'] = moc.origin
        tbhdu.header.comments['ORIGIN'] = 'MOC origin'
    tbhdu.header['DATE'] = datetime.utcnow().replace(
        microsecond=0).isoformat()
    tbhdu.header.comments['DATE'] = 'MOC creation date'
    if moc.name is not None:
        tbhdu.header['EXTNAME'] = moc.name
        tbhdu.header.comments['EXTNAME'] = 'MOC name'

    return tbhdu


def write_moc_fits(moc, filename, **kwargs):
    """Write a MOC as a FITS file.

    Any additional keyword arguments are passed to the
    astropy.io.fits.HDUList.writeto method.
    """

    tbhdu = write_moc_fits_hdu(moc)
    prihdr = fits.Header()
    prihdu = fits.PrimaryHDU(header=prihdr)
    hdulist = fits.HDUList([prihdu, tbhdu])
    hdulist.writeto(filename, **kwargs)


def read_moc_fits(moc, filename, include_meta=False, **kwargs):
    """Read data from a FITS file into a MOC.

    Any additional keyword arguments are passed to the
    astropy.io.fits.open method.
    """

    hl = fits.open(filename, mode='readonly', **kwargs)

    read_moc_fits_hdu(moc, hl[1], include_meta)


def read_moc_fits_hdu(moc, hdu, include_meta=False):
    """Read data from a FITS table HDU into a MOC.
    """

    if include_meta:
        header = hdu.header

        if 'MOCTYPE' in header:
            moc.type = header['MOCTYPE']
        if 'MOCID' in header:
            moc.id = header['MOCID']
        if 'ORIGIN' in header:
            moc.origin = header['ORIGIN']
        if 'EXTNAME' in header:
            moc.name = header['EXTNAME']

    current_order = None
    current_cells = []

    # Determine type to use for orders: 32 bit if column type is J,
    # otherwise assume we need 64 bit.
    moc_type = np.int32 if (hdu.data.formats[0] == 'J') else np.int64

    nuniqs = hdu.data.field(0)
    orders = (np.log2(nuniqs / 4) / 2).astype(moc_type)
    cells = nuniqs - 4 * (4 ** orders)

    for (order, cell) in izip(orders, cells):
        if order != current_order:
            if current_cells:
                moc.add(current_order, current_cells)

            current_order = order
            current_cells = [cell]

        else:
            current_cells.append(cell)

    if current_cells:
        moc.add(current_order, current_cells)



# Copyright (C) 2014 Science and Technology Facilities Council.
# Copyright (C) 2017-2024 East Asian Observatory.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from io import BytesIO
from unittest import TestCase

from pymoc import MOC
from pymoc.io.json import read_moc_json, write_moc_json


class JSONTestCase(TestCase):
    def test_json(self):
        test_json = b'{"1":[1,2,4],"2":[12,13,14,21,23,25]}'
        in_ = BytesIO(test_json)

        moc = MOC()
        read_moc_json(moc, file=in_)

        self.assertEqual(moc.order, 2)
        self.assertEqual(moc[0], frozenset())
        self.assertEqual(moc[1], frozenset([1, 2, 4]))
        self.assertEqual(moc[2], frozenset([12, 13, 14, 21, 23, 25]))

        out = BytesIO()
        write_moc_json(moc, file=out)

        self.assertEqual(out.getvalue(), test_json)

    def test_json_trailing(self):
        # Check MOC 1.1 addition of trailing section to
        # signify the MOCORDER.
        in_ = BytesIO(b'{"13":[5,6,7],"14":[]}')

        moc = MOC()
        read_moc_json(moc, file=in_)

        self.assertEqual(moc[13], frozenset([5, 6, 7]))
        self.assertEqual(moc[14], frozenset())

    def test_json_large(self):
        orig = MOC()
        orig.add(29, [
            3458700000000000000, 3458700000000000007,
            3458700000000000008, 3458700000000000009,
        ])

        out = BytesIO()
        write_moc_json(orig, file=out)
        json = out.getvalue()

        self.assertEqual(
            json, b'{"29":[3458700000000000000,3458700000000000007,'
            b'3458700000000000008,3458700000000000009]}')

        copy = MOC()
        in_ = BytesIO(json)
        read_moc_json(copy, file=in_)

        self.assertEqual(copy.order, 29)
        self.assertEqual(copy.cells, 4)
        self.assertEqual(copy[29], frozenset([
            3458700000000000000, 3458700000000000007,
            3458700000000000008, 3458700000000000009]))



# Copyright (C) 2015-2017 East Asian Observatory.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

from math import pi

from astropy.coordinates import SkyCoord
from astropy.io import ascii
from astropy.units import arcsecond, hour, degree, radian
from astropy.units.quantity import Quantity
from healpy import query_disc
from healpy.pixelfunc import \
    ang2pix, ang2vec, get_all_neighbours, pix2ang, vec2pix
import numpy as np

from ..moc import MOC


def catalog_to_moc(catalog, radius, order, **kwargs):
    """
    Convert a catalog to a MOC.

    The catalog is given as an Astropy SkyCoord object containing
    multiple coordinates.  The radius of catalog entries can be
    given as an Astropy Quantity (with units), otherwise it is assumed
    to be in arcseconds.

    Any additional keyword arguments are passed on to `catalog_to_cells`.
    """

    # Generate list of MOC cells.
    cells = catalog_to_cells(catalog, radius, order, **kwargs)

    # Create new MOC object using our collection of cells.
    moc = MOC(moctype='CATALOG')
    moc.add(order, cells, no_validation=True)
    return moc


def _catalog_to_cells_neighbor(catalog, radius, order):
    """
    Convert a catalog to a list of cells.

    This is the original implementation of the `catalog_to_cells`
    function which does not make use of the Healpy `query_disc` routine.

    Note: this function uses a simple flood-filling approach and is
    very slow, especially when used with a large radius for catalog objects
    or a high resolution order.
    """

    if not isinstance(radius, Quantity):
        radius = radius * arcsecond

    nside = 2 ** order

    # Ensure catalog is in ICRS coordinates.
    catalog = catalog.icrs

    # Determine central cell for each catalog entry.
    phi = catalog.ra.radian
    theta = (pi / 2) - catalog.dec.radian

    cells = np.unique(ang2pix(nside, theta, phi, nest=True))

    # Iteratively consider the neighbors of cells within our
    # catalog regions.
    new_cells = cells
    rejected = np.array((), dtype=np.int64)
    while True:
        # Find new valid neighboring cells which we didn't already
        # consider.
        neighbors = np.unique(np.ravel(
            get_all_neighbours(nside, new_cells, nest=True)))

        neighbors = np.extract(
            [(x != -1) and (x not in cells) and (x not in rejected)
             for x in neighbors], neighbors)

        # Get the coordinates of each of these neighbors and compare them
        # to the catalog entries.
        (theta, phi) = pix2ang(nside, neighbors, nest=True)

        coords = SkyCoord(phi, (pi / 2) - theta, frame='icrs', unit='rad')

        (idx, sep2d, dist3d) = coords.match_to_catalog_sky(catalog)

        within_range = (sep2d < radius)

        # If we didn't find any new cells within range,
        # end the iterative process.
        if not np.any(within_range):
            break

        new_cells = neighbors[within_range]
        cells = np.concatenate((cells, new_cells))
        rejected = np.concatenate((
            rejected, neighbors[np.logical_not(within_range)]))

    return cells


def catalog_to_cells(catalog, radius, order, include_fallback=True, **kwargs):
    """
    Convert a catalog to a set of cells.

    This function is intended to be used via `catalog_to_moc` but
    is available for separate usage.  It takes the same arguments
    as that function.

    This function uses the Healpy `query_disc` function to get a list
    of cells for each item in the catalog in turn.  Additional keyword
    arguments, if specified, are passed to `query_disc`.  This can include,
    for example, `inclusive` (set to `True` to include cells overlapping
    the radius as well as those with centers within it) and `fact`
    (to control sampling when `inclusive` is specified).

    If cells at the given order are bigger than the given radius, then
    `query_disc` may find none inside the radius.  In this case,
    if `include_fallback` is `True` (the default), the cell at each
    position is included.

    If the given radius is zero (or smaller) then Healpy `query_disc`
    is not used -- instead the fallback position is used automatically.
    """

    nside = 2 ** order

    # Ensure catalog is in ICRS coordinates.
    catalog = catalog.icrs

    # Ensure radius is in radians.
    if isinstance(radius, Quantity):
        radius = radius.to(radian).value
    else:
        radius = radius * pi / (180.0 * 3600.0)

    # Convert coordinates to position vectors.
    phi = catalog.ra.radian
    theta = (pi / 2) - catalog.dec.radian

    vectors = ang2vec(theta, phi)

    # Ensure we can iterate over vectors (it might be a single position).
    if catalog.isscalar:
        vectors = [vectors]

    # Query for a list of cells for each catalog position.
    cells = set()
    for vector in vectors:
        if radius > 0.0:
            # Try "disc" query.
            vector_cells = query_disc(nside, vector, radius, nest=True, **kwargs)

            if vector_cells.size > 0:
                cells.update(vector_cells.tolist())
                continue

            elif not include_fallback:
                continue

        # The query didn't find anything -- include the cell at the
        # given position at least.
        cell = vec2pix(nside, vector[0], vector[1], vector[2], nest=True)
        cells.add(cell.item())

    return cells


def read_ascii_catalog(filename, format_, unit=None):
    """
    Read an ASCII catalog file using Astropy.

    This routine is used by pymoctool to load coordinates from a
    catalog file in order to generate a MOC representation.
    """

    catalog = ascii.read(filename, format=format_)
    columns = catalog.columns

    if 'RA' in columns and 'Dec' in columns:
        if unit is None:
            unit = (hour, degree)

        coords = SkyCoord(catalog['RA'],
                          catalog['Dec'],
                          unit=unit,
                          frame='icrs')

    elif 'Lat' in columns and 'Lon' in columns:
        if unit is None:
            unit = (degree, degree)

        coords = SkyCoord(catalog['Lon'],
                          catalog['Lat'],
                          unit=unit,
                          frame='galactic')

    else:
        raise Exception('columns RA,Dec or Lon,Lat not found')

    return coords



# Copyright (C) 2016 East Asian Observatory.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from random import sample
from unittest import TestCase

from pymoc import MOC


class OperatorsRandomizedTestCase(TestCase):
    def test_operators(self):
        repeats = 10
        order = 6

        for i in range(0, repeats):
            # Make random MOCs, take difference and intersection.
            p = self._make_random_moc(order)
            q = self._make_random_moc(order)

            s_pq = p - q
            s_qp = q - p
            i_pq = p.intersection(q)
            i_qp = q.intersection(p)

            # p.write('test_moc_p_{}.fits'.format(i + 1), overwrite=True)
            # q.write('test_moc_q_{}.fits'.format(i + 1), overwrite=True)
            # s_pq.write('test_moc_pq_{}.fits'.format(i + 1), overwrite=True)
            # s_qp.write('test_moc_qp_{}.fits'.format(i + 1), overwrite=True)
            # i_pq.write('test_moc_i_{}.fits'.format(i + 1), overwrite=True)

            # Intersection should be the same both ways round.
            self.assertEqual(i_pq, i_qp)

            # Areas should add up to differences + 2 x intersection.
            self.assertAlmostEqual(p.area + q.area,
                                   s_pq.area + s_qp.area + 2 * i_pq.area,
                                   places=10)

            # Make differences and intersection via flattened set.
            p_flat = p.flattened(order=order)
            q_flat = q.flattened(order=order)

            s_pq_flat = p_flat - q_flat
            s_qp_flat = q_flat - p_flat
            i_flat = p_flat & q_flat

            # Compare with flattened-method version via MOC equality.
            self.assertEqual(s_pq, MOC(order, s_pq_flat))
            self.assertEqual(s_qp, MOC(order, s_qp_flat))
            self.assertEqual(i_pq, MOC(order, i_flat))

            # Compare with flattened-method version via flat set.
            s_pq_flattened = s_pq.flattened(order=order)
            s_qp_flattened = s_qp.flattened(order=order)
            i_pq_flattened = i_pq.flattened(order=order)
            i_qp_flattened = i_qp.flattened(order=order)

            self.assertEqual(s_pq_flattened, s_pq_flat)
            self.assertEqual(s_qp_flattened, s_qp_flat)
            self.assertEqual(i_pq_flattened, i_flat)
            self.assertEqual(i_qp_flattened, i_flat)

            # Differences and intersection should be disjoint.
            self.assertTrue(s_pq_flattened.isdisjoint(s_qp_flattened))
            self.assertTrue(s_pq_flattened.isdisjoint(i_pq_flattened))
            self.assertTrue(s_qp_flattened.isdisjoint(i_pq_flattened))

    def _make_random_moc(self, order):
        m = MOC()

        for order_i in range(0, order + 1):
            m.add(order_i, sample(range(0, 12 * 4 ** order_i), 4 ** order_i))

        return m



# Copyright (C) 2014 Science and Technology Facilities Council.
# Copyright (C) 2017 East Asian Observatory.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals


def write_moc_ascii(moc, filename=None, file=None):
    """Write a MOC to an ASCII file.

    Either a filename, or an open file object can be specified.
    """

    orders = []

    for (order, cells) in moc:
        ranges = []
        rmin = rmax = None

        for cell in sorted(cells):
            if rmin is None:
                rmin = rmax = cell
            elif rmax == cell - 1:
                rmax = cell
            else:
                ranges.append(_format_range(rmin, rmax))
                rmin = rmax = cell

        ranges.append(_format_range(rmin, rmax))

        orders.append('{0}'.format(order) + '/' + ','.join(ranges))

    if file is not None:
        _write_ascii(orders, file)
    else:
        with open(filename, 'w') as f:
            _write_ascii(orders, f)


def read_moc_ascii(moc, filename=None, file=None):
    """Read from an ASCII file into a MOC.

    Either a filename, or an open file object can be specified.
    """

    if file is not None:
        orders = _read_ascii(file)
    else:
        with open(filename, 'r') as f:
            orders = _read_ascii(f)

    for text in orders:
        if not text:
            continue

        cells = []
        (order, ranges) = text.split('/')

        if not ranges:
            continue

        for r in ranges.split(','):
            try:
                cells.append(int(r))
            except ValueError as e:
                (rmin, rmax) = r.split('-')
                cells.extend(range(int(rmin), int(rmax) + 1))

        moc.add(order, cells)


def _write_ascii(orders, f):
    f.write(' '.join(orders))


def _read_ascii(f):
    text = f.read()

    return text.strip().split(' ')


def _format_range(rmin, rmax):
    if rmin == rmax:
        return '{0}'.format(rmin)
    elif rmax == rmin + 1:
        return '{0},{1}'.format(rmin, rmax)
    else:
        return '{0}-{1}'.format(rmin, rmax)



# Copyright (C) 2014 Science and Technology Facilities Council.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, unicode_literals

from codecs import utf_8_decode, utf_8_encode
import json


def write_moc_json(moc, filename=None, file=None):
    """Write a MOC in JSON encoding.

    Either a filename, or an open file object can be specified.
    """

    moc.normalize()

    obj = {}

    for (order, cells) in moc:
        obj['{0}'.format(order)] = sorted(cells)

    if file is not None:
        _write_json(obj, file)
    else:
        with open(filename, 'wb') as f:
            _write_json(obj, f)


def read_moc_json(moc, filename=None, file=None):
    """Read JSON encoded data into a MOC.

    Either a filename, or an open file object can be specified.
    """

    if file is not None:
        obj = _read_json(file)
    else:
        with open(filename, 'rb') as f:
            obj = _read_json(f)

    for (order, cells) in obj.items():
        moc.add(order, cells)


def _write_json(obj, f):
    f.write(utf_8_encode(
        json.dumps(obj, sort_keys=True, separators=(',', ':')))[0])


def _read_json(f):
    return json.loads(utf_8_decode(f.read())[0])



# Copyright (C) 2013-2014 Science and Technology Facilities Council.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import healpy
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.pyplot as plt
import numpy as np


def plot_moc(moc, order=None, antialias=0, filename=None,
             projection='cart', color='blue', title='', coord_sys='C',
             graticule=True, **kwargs):
    """Plot a MOC using Healpy.

    This generates a plot of the MOC at the specified order, or the MOC's
    current order if this is not specified.  The MOC is flattened at an order
    of `order + antialias` to generate intermediate color levels.

    :param order: HEALPix order at which to generate the plot.

    :param antialias: number of additional HEALPix orders to use for
        intermediate color levels.  (There can be `4 ** antialias` levels.)

    :param filename: file in which to save plot.  If not specified then
        the plot is shown with `plt.show()`.

    :param projection: map projection to be used --- can be shortened to
        4 characters.  One of:

            * `'cart[esian]'` (uses `healpy.visufunc.cartview`)
            * `'moll[weide]'` (uses `healpy.visufunc.mollview`)
            * `'gnom[onic]'` (uses `healpy.visufunc.gnomview`)

    :param color: color scheme.
        One of:

            * `'blue'`
            * `'green'`
            * `'red'`
            * `'black'`

    :param title: title of the plot.

    :param coord_sys: Healpy coordinate system code for the desired plot
        coordinates.  One of:

            * `'C'` --- Celestial (equatorial)
            * `'G'` --- Galactic
            * `'E'` --- Ecliptic

    :param graticule: whether or not to draw a graticule.

    :param \*\*kwargs: passed to the selected Healpy plotting function.
    """

    # Process arguments.
    plotargs = {'xsize': 3200, 'cbar': False, 'notext': True}

    if order is None:
        order = moc.order

    if projection.startswith('cart'):
        plotter = healpy.visufunc.cartview
    elif projection.startswith('moll'):
        plotter = healpy.visufunc.mollview
    elif projection.startswith('gnom'):
        plotter = healpy.visufunc.gnomview
    else:
        raise ValueError('Unknown projection: {0}'.format(projection))

    if color == 'blue':
        plotargs['cmap'] = LinearSegmentedColormap.from_list(
            'white-blue', ['#FFFFFF', '#0000AA'])
    elif color == 'green':
        plotargs['cmap'] = LinearSegmentedColormap.from_list(
            'white-green', ['#FFFFFF', '#008800'])
    elif color == 'red':
        plotargs['cmap'] = LinearSegmentedColormap.from_list(
            'white-red', ['#FFFFFF', '#FF0000'])
    elif color == 'black':
        plotargs['cmap'] = LinearSegmentedColormap.from_list(
            'white-black', ['#FFFFFF', '#000000'])
    else:
        raise ValueError('Unknown color: {0}'.format(color))

    if coord_sys == 'C':
        pass
    elif coord_sys == 'G':
        plotargs['coord'] = ('C', 'G')
    elif coord_sys == 'E':
        plotargs['coord'] = ('C', 'E')
    else:
        raise ValueError('Unknown coordinate system: {0}'.format(coord_sys))

    # Any other arguments are passed the Healpy plotter directly.
    plotargs.update(kwargs)

    # Create a Numpy array which is zero for points outside the MOC and one
    # for points inside the MOC.
    map = np.zeros(12 * 4 ** order)
    antialias_shift = 2 * antialias

    for cell in moc.flattened(order + antialias):
        map[cell >> antialias_shift] += 1.0

    # Plot the Numpy array using Healpy.
    plotter(map, nest=True, title=title, **plotargs)

    if graticule:
        healpy.visufunc.graticule()

    if filename is not None:
        plt.savefig(filename)
    else:
        plt.show()



# Copyright (C) 2014 Science and Technology Facilities Council.
# Copyright (C) 2017-2024 East Asian Observatory.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from io import StringIO
from unittest import TestCase

from pymoc import MOC
from pymoc.io.ascii import read_moc_ascii, write_moc_ascii


class ASCIITestCase(TestCase):
    def test_ascii(self):
        test_ascii = '1/1,3,4 2/4,25,12-14,21'
        test_ascii_sorted = '1/1,3,4 2/4,12-14,21,25'
        in_ = StringIO(test_ascii)

        moc = MOC()
        read_moc_ascii(moc, file=in_)

        self.assertEqual(moc.order, 2)
        self.assertEqual(moc[0], frozenset())
        self.assertEqual(moc[1], frozenset([1, 3, 4]))
        self.assertEqual(moc[2], frozenset([4, 12, 13, 14, 21, 25]))

        out = StringIO()
        write_moc_ascii(moc, file=out)

        self.assertEqual(out.getvalue(), test_ascii_sorted)

    def test_ascii_empty(self):
        in_ = StringIO('')

        moc = MOC()
        read_moc_ascii(moc, file=in_)

        self.assertEqual(moc.order, 0)
        self.assertEqual(moc[0], frozenset())

    def test_ascii_trailing(self):
        # Check MOC 1.1 addition of trailing section to
        # signify the MOCORDER.
        in_ = StringIO('13/5,6,7 14/')

        moc = MOC()
        read_moc_ascii(moc, file=in_)

        self.assertEqual(moc[13], frozenset([5, 6, 7]))
        self.assertEqual(moc[14], frozenset())

    def test_ascii_large(self):
        orig = MOC()
        orig.add(29, [
            3458700000000000000, 3458700000000000007,
            3458700000000000008, 3458700000000000009,
        ])

        out = StringIO()
        write_moc_ascii(orig, file=out)
        text = out.getvalue()

        self.assertEqual(
            text, '29/3458700000000000000,'
            '3458700000000000007-3458700000000000009')

        copy = MOC()
        in_ = StringIO(text)
        read_moc_ascii(copy, file=in_)

        self.assertEqual(copy.order, 29)
        self.assertEqual(copy.cells, 4)
        self.assertEqual(copy[29], frozenset([
            3458700000000000000, 3458700000000000007,
            3458700000000000008, 3458700000000000009]))



# Copyright (C) 2017 East Asian Observatory.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import TestCase

from astropy.coordinates import SkyCoord

from pymoc import MOC
from pymoc.util.catalog import catalog_to_moc, catalog_to_cells


class CatalogTestCase(TestCase):
    def test_cells(self):
        # Catalog with an entry at cell 30000 (order 6: ~1 deg. res.)
        catalog = SkyCoord(303.75, -4.18152827, frame='icrs', unit='deg')

        cells = catalog_to_cells(catalog, 60, 6, include_fallback=False)
        self.assertEqual(cells, set((30000,)))

        # Nudge position so it no longer matches: no result without fallback.
        catalog = SkyCoord(303.50, -4.18152827, frame='icrs', unit='deg')

        cells = catalog_to_cells(catalog, 60, 6, include_fallback=False)
        self.assertEqual(cells, set(()))

        cells = catalog_to_cells(catalog, 60, 6, include_fallback=True)
        self.assertEqual(cells, set((30000,)))

        # Catalog with two entries: compare to results from flood-fill method.
        catalog = SkyCoord([100.0, 200.0], [40.0, 60.0],
                           frame='icrs', unit='deg')

        cells = catalog_to_cells(catalog, 3600, 7)
        self.assertEqual(cells, set((
            26860, 26861, 26862, 26863, 26866, 26867, 26872, 26873,
            26874, 26875, 27204, 27205, 27206, 27207, 27216,
            44535, 44540, 44541, 44543, 47264, 47265, 47266, 47267,
            47268, 47270, 47271, 47272, 47273, 47274, 47276,
        )))

        # Test inclusive option: should add extra cells.
        inclusive = catalog_to_cells(catalog, 3600, 7, inclusive=True)
        self.assertGreater(len(inclusive), len(cells))
        self.assertTrue(inclusive.issuperset(cells))

        # Test a cell close to a boundary.
        catalog = SkyCoord(44.978, 1.304, frame='icrs', unit='deg')
        cells = catalog_to_cells(catalog, 10, 12, inclusive=True)
        self.assertEqual(cells, set((12344, 12345, 12346, 12347,)))

        cells = catalog_to_cells(catalog, 0, 12, inclusive=True)
        self.assertEqual(cells, set((12344,)))

    def test_catalog(self):
        catalog = SkyCoord([150.0, 300.0], [-45.0, 45.0],
                           frame='icrs', unit='deg')

        moc = catalog_to_moc(catalog, 0.5, 20)

        self.assertIsInstance(moc, MOC)
        self.assertEqual(moc.type, 'CATALOG')

        expected = MOC()
        expected.add(19, [
            994284930659, 994284930660, 994284930662, 2579127859609,
            2579127859611, 2579127859612])
        expected.add(20, [
            3977139722630, 3977139722631, 3977139722644, 3977139722646,
            3977139722652, 3977139722661, 10316511438426, 10316511438435,
            10316511438441, 10316511438443, 10316511438456, 10316511438457])

        self.assertEqual(moc, expected)



# Copyright (C) 2014 Science and Technology Facilities Council.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from math import sqrt
from unittest import TestCase

from pymoc import MOC


class AreaTestCase(TestCase):
    def test_orders(self):
        # Test the size of one cell of each order against the values
        # given in the MOC recommendation.

        m = MOC(0, [0])
        self.assertAlmostEqual(sqrt(m.area_sq_deg), 58.63, places=2)
        m = MOC(1, [0])
        self.assertAlmostEqual(sqrt(m.area_sq_deg), 29.32, places=2)
        m = MOC(2, [0])
        self.assertAlmostEqual(sqrt(m.area_sq_deg), 14.66, places=2)
        m = MOC(3, [0])
        self.assertAlmostEqual(sqrt(m.area_sq_deg), 7.329, places=3)
        m = MOC(4, [0])
        self.assertAlmostEqual(sqrt(m.area_sq_deg), 3.665, places=3)
        m = MOC(5, [0])
        self.assertAlmostEqual(sqrt(m.area_sq_deg), 1.832, places=3)
        m = MOC(6, [0])
        self.assertAlmostEqual(60 * sqrt(m.area_sq_deg), 54.97, places=2)
        m = MOC(7, [0])
        self.assertAlmostEqual(60 * sqrt(m.area_sq_deg), 27.48, places=2)
        m = MOC(8, [0])
        self.assertAlmostEqual(60 * sqrt(m.area_sq_deg), 13.74, places=2)
        m = MOC(9, [0])
        self.assertAlmostEqual(60 * sqrt(m.area_sq_deg), 6.871, places=3)
        m = MOC(10, [0])
        self.assertAlmostEqual(60 * sqrt(m.area_sq_deg), 3.435, places=3)
        m = MOC(11, [0])
        self.assertAlmostEqual(60 * sqrt(m.area_sq_deg), 1.718, places=3)
        m = MOC(12, [0])
        self.assertAlmostEqual(3600 * sqrt(m.area_sq_deg), 51.53, places=2)
        m = MOC(13, [0])
        self.assertAlmostEqual(3600 * sqrt(m.area_sq_deg), 25.77, places=2)
        m = MOC(14, [0])
        self.assertAlmostEqual(3600 * sqrt(m.area_sq_deg), 12.88, places=2)
        m = MOC(15, [0])
        self.assertAlmostEqual(3600 * sqrt(m.area_sq_deg), 6.442, places=3)
        m = MOC(16, [0])
        self.assertAlmostEqual(3600 * sqrt(m.area_sq_deg), 3.221, places=3)
        m = MOC(17, [0])
        self.assertAlmostEqual(3600 * sqrt(m.area_sq_deg), 1.610, places=3)
        m = MOC(18, [0])
        self.assertAlmostEqual(3600E3 * sqrt(m.area_sq_deg), 805.2, places=1)
        m = MOC(19, [0])
        self.assertAlmostEqual(3600E3 * sqrt(m.area_sq_deg), 402.6, places=1)
        m = MOC(20, [0])
        self.assertAlmostEqual(3600E3 * sqrt(m.area_sq_deg), 201.3, places=1)
        m = MOC(21, [0])
        self.assertAlmostEqual(3600E3 * sqrt(m.area_sq_deg), 100.6, places=1)
        m = MOC(22, [0])
        self.assertAlmostEqual(3600E3 * sqrt(m.area_sq_deg), 50.32, places=2)
        m = MOC(23, [0])
        self.assertAlmostEqual(3600E3 * sqrt(m.area_sq_deg), 25.16, places=2)
        m = MOC(24, [0])
        self.assertAlmostEqual(3600E3 * sqrt(m.area_sq_deg), 12.58, places=2)
        m = MOC(25, [0])
        self.assertAlmostEqual(3600E3 * sqrt(m.area_sq_deg), 6.291, places=3)
        m = MOC(26, [0])
        self.assertAlmostEqual(3600E3 * sqrt(m.area_sq_deg), 3.145, places=3)
        m = MOC(27, [0])
        self.assertAlmostEqual(3600E3 * sqrt(m.area_sq_deg), 1.573, places=3)
        m = MOC(28, [0])
        self.assertAlmostEqual(3600E6 * sqrt(m.area_sq_deg), 786.3, places=1)
        m = MOC(29, [0])
        self.assertAlmostEqual(3600E6 * sqrt(m.area_sq_deg), 393.2, places=1)
